// The sample box model.  You should build a file
// very similar to this for when you make your model in
// order to plug in to the animator project.

#pragma warning (disable : 4305)
#pragma warning (disable : 4244)
#pragma warning (disable : 4786)
#pragma warning (disable : 4312)

#include "modelerview.h"
#include "modelerdraw.h"
#include "modelerui.h"
#include "vault.h"
#include "FL/glut.h"
#include "model.h"
#include "cameras.h"
#include <cmath>
#include <time.h>

/** The scene, which includes the lights and models. */
class Scene : public Model {
protected:
///////////////////////////////// TEXTURES ////////////////////////////////////
	Texture2D floorTexture;


///////////////////////////////// SHADERS /////////////////////////////////////
	// ShaderProgram shader;

//////////////////////////////// PROPERTIES ///////////////////////////////////

	// Scene lights
	PointLight pointLight;
	DirectionalLight directionalLight;

	// A robot arm model
	CuriosityRover rover;

public:
	ParticleSystem ps;

	/** Modeler calls this to get our particle system. */
	ParticleSystem* getParticleSystem() { return rover.getParticleSystem(); }

	/** Construct the scene */
	Scene() :
	    // You have to call the parent class's constructor, to provide a
	    // name for the model.
	    Model("Scene"),

		// Construct textures and shaders. 
		// They won't be loaded until the model is drawn for the first time.
		floorTexture("mars_surface2.png"),
		// shader("shader.vert", "shader.frag", NULL),
		rover("Curiosity", 0.0, 0.0, 0.0, /*size*/ 2.0),

		// Call the constructors for the lights
		pointLight("Point Light", GL_LIGHT1, /**direction part**/ -5, 5, 5, /**diffuse part**/ 1.0, 0.5, 0.5, 
		/**specular part**/ 1.0, 0.5, 0.5, /**ambient part**/ .2f, 0.1, 0.1 /**attenuation part**/, 0.4, 0.7, 0),
		directionalLight("Directional Light", GL_LIGHT0, /**direction part**/ 5, 5, 5, /**diffuse part**/ 0.0f, 0.5, 0.5f, 
		/**specular part**/ 0.0f, 0.5f, 0.5f )

		// Now, call the constructors for each Property:
		{
		// If you have child Models, like the MobileLight model from model.h,
		// you can add their property groups, and they will appear in the list
		// in the top left corner of Modeler, under this model's entry:
		properties.add(pointLight.getProperties())
				  .add(directionalLight.getProperties());
		properties.add(rover.getProperties());


		// Finally, add all the properties to this model's PropertyGroup.
		}

	/**
	 * Modeler calls this method once an OpenGL context becomes available,
	 * (which happens before the first time draw() is called), and whenever
	 * you click the "Reload Textures and Shaders" menu option.
	 *
	 * Make sure you call the load() methods of each of your textures and
	 * shaders here.
	 */
	void load() {
		floorTexture.load();
		rover.load();
		// shader.load();
	}

	/**
	 * Modeler calls this method many times a second when the Animate
	 * checkbox is checked.
	 */
	void tick() {
		// You can put code here to increment animation counters for
		// extra credit.
	}


	/** Draw the scene. */
	void draw() {
		// Use supplied shader
		// shader.use();

		// Draw floor with texture
		floorTexture.use();

		// Draw floor as a big circle
		float radius = 100.f;
		float steps = 100;
		float angle_step = 2.f * M_PI / steps;
		glBegin(GL_TRIANGLES);
			for (float angle = 0; angle < 2.f * M_PI; angle += angle_step)
			{
				float x, z;
				
				// Center vertex
				x = 0.f;
				z = 0.f;
				
				glTexCoord2f(0.5f, 0.5f); // specify the texture coordinate
				glNormal3f(0, 1, 0); // specify the surface's normal at this vertex
				glVertex3f(x, -2, z); // both before its corresponding vertex

				// First outer vertex
				x = sin(angle) * radius;
				z = cos(angle) * radius;
				
				glTexCoord2f(0.5f * sin(angle) + 0.5f, 0.5f * cos(angle) + 0.5f); // specify the texture coordinate
				glNormal3f(0, 1, 0); // specify the surface's normal at this vertex
				glVertex3f(x, -2, z); // both before its corresponding vertex

				// Second outer vertex (increased angle)
				x = sin(angle + angle_step) * radius;
				z = cos(angle + angle_step) * radius;
				
				glTexCoord2f(0.5f * sin(angle + angle_step) + 0.5f, 0.5f * cos(angle + angle_step) + 0.5f); // specify the texture coordinate
				glNormal3f(0, 1, 0); // specify the surface's normal at this vertex
				glVertex3f(x, -2, z); // both before its corresponding vertex
			}
		glEnd();

		// The lights must be drawn FIRST, so the other scene elements
		// can get lit!
		pointLight.draw();
		directionalLight.draw();
		


		// Draw Curiosity Rover
		rover.draw();
	}
};

/**
 * The program starts here.
 */
int main()
{
	// Tell the FLTK library to let us use OpenGL
	Fl::gl_visual(FL_RGB);

	//init_load_curve_file();

	srand(time(0));

	// Instantiate Modeler
	ModelerUserInterface ui;

	// Give Modeler your scene.
	// Modeler will free this pointer when it exits.
	ui.setModel(new Scene());


	// Run the application
    return ui.run();
}